from torch import nn


class Identity(nn.Identity):
    """
    nn.Identity but ignores all passed arguments
    useful for replacing normalization layers with identity without worrying about the passed arguments
    Examples:
        >>> if norm_mode == "batchnorm":
        >>>   norm_ctor = nn.BatchNorm1d
        >>> elif norm_mode == "none":
        >>>   norm_ctor = Identity
        >>> else:
        >>>   raise NotImplemented
        >>> ...
        >>> self.norm = norm_ctor(dim)
    """

    def __init__(self, *_, **__):
        super().__init__()



from torch import nn
from kappamodules.utils.param_checking import to_2tuple

class VitMlp(nn.Module):
    def __init__(
            self,
            in_dim,
            hidden_dim=None,
            out_dim=None,
            act_ctor=nn.GELU,
            bias=True,
    ):
        super().__init__()
        out_dim = out_dim or in_dim
        hidden_dim = hidden_dim or in_dim
        bias1, bias2 = to_2tuple(bias)

        self.fc1 = nn.Linear(in_dim, hidden_dim, bias=bias1)
        self.act = act_ctor()
        self.fc2 = nn.Linear(hidden_dim, out_dim, bias=bias2)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.fc2(x)
        return x



import math

import torch
import torch.nn.functional as F
from torch import nn


class WeightNormLinear(nn.Module):
    """
    torch.nn.utils.weight_norm(nn.Linear(...)) but with weight_g as buffer when it is fixed
    if weight_g is set to requires_grad=False (as done in DINO, iBOT, MUGS, ...) and during
    training a parent module is unfrozen via
    ```
    for p in module.parameters():
      p.requires_grad = True
    ```
    the weight_g is also unfrozen. registering weight_g as a buffer avoids this
    """

    def __init__(
            self,
            in_features: int,
            out_features: int,
            bias: bool = True,
            fixed_g: bool = False,
            device=None,
            dtype=None,
    ):
        super().__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        self.in_features = in_features
        self.out_features = out_features
        self.fixed_g = fixed_g
        self.weight_v = nn.Parameter(torch.empty(out_features, in_features, **factory_kwargs))
        if fixed_g:
            self.register_buffer("weight_g", torch.empty(out_features, 1, **factory_kwargs))
        else:
            self.weight_g = nn.Parameter(torch.empty(out_features, 1, **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty(out_features, **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        torch.nn.init.kaiming_uniform_(self.weight_v, a=math.sqrt(5))
        with torch.no_grad():
            if self.fixed_g:
                self.weight_g.fill_(1)
            else:
                self.weight_g.copy_(torch.norm_except_dim(self.weight_v))
        if self.bias is not None:
            # noinspection PyProtectedMember
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight_v)
            bound = 1 / math.sqrt(fan_in)
            torch.nn.init.uniform_(self.bias, -bound, bound)

    def extra_repr(self) -> str:
        return nn.Linear.extra_repr(self)

    def forward(self, x):
        # noinspection PyProtectedMember
        return F.linear(x, torch._weight_norm(self.weight_v, self.weight_g, 0), self.bias)



import unittest

import torch
from timm.layers import Mlp

from kappamodules.vit import VitMlp


class TestVitMlp(unittest.TestCase):
    def test_equal_to_timm(self):
        seed = 89453
        torch.manual_seed(seed)
        mlp0 = Mlp(in_features=4, hidden_features=12)
        torch.manual_seed(seed)
        mlp1 = VitMlp(in_dim=4, hidden_dim=12)
        torch.manual_seed(seed)
        x = torch.randn(3, 4)
        y0 = mlp0(x)
        y1 = mlp1(x)
        self.assertTrue(torch.all(y0 == y1))



import unittest

import torch
from torch import nn

from kappamodules.layers import Residual


class TestResidual(unittest.TestCase):
    def test_single_module(self):
        # original
        torch.manual_seed(0)
        og_x = torch.randn(2, 4)
        og_layer = nn.Linear(4, 4)
        og_y = og_x + og_layer(og_x)
        og_y.mean().backward()

        # Residual
        torch.manual_seed(0)
        res_x = torch.randn(2, 4)
        res_layer = nn.Linear(4, 4)
        res_module = Residual(res_layer)
        res_y = res_module(res_x)
        res_y.mean().backward()

        # checks
        self.assertTrue(torch.all(og_y == res_y))
        self.assertTrue(torch.all(og_layer.weight.grad == res_layer.weight.grad))
        self.assertTrue(torch.all(og_layer.bias.grad == res_layer.bias.grad))

    def test_expand_module(self):
        # original
        torch.manual_seed(0)
        og_x = torch.randn(2, 4)
        og_layer0 = nn.Linear(4, 8)
        og_layer1 = nn.Linear(8, 4)
        og_layer = nn.Sequential(og_layer0, og_layer1)
        og_y = og_x + og_layer(og_x)
        og_y.mean().backward()

        # Residual
        torch.manual_seed(0)
        res_x = torch.randn(2, 4)
        res_layer0 = nn.Linear(4, 8)
        res_layer1 = nn.Linear(8, 4)
        res_module = Residual(res_layer0, res_layer1)
        res_y = res_module(res_x)
        res_y.mean().backward()

        # checks
        self.assertTrue(torch.all(og_y == res_y))
        self.assertTrue(torch.all(og_layer0.weight.grad == res_layer0.weight.grad))
        self.assertTrue(torch.all(og_layer0.bias.grad == res_layer0.bias.grad))
        self.assertTrue(torch.all(og_layer1.weight.grad == res_layer1.weight.grad))
        self.assertTrue(torch.all(og_layer1.bias.grad == res_layer1.bias.grad))



from torch import nn

# region layers
ALL_BATCHNORMS = [
    nn.BatchNorm1d,
    nn.BatchNorm2d,
    nn.BatchNorm3d,
    nn.LazyBatchNorm1d,
    nn.LazyBatchNorm2d,
    nn.LazyBatchNorm3d,
    nn.SyncBatchNorm,
]

ALL_NORMS = [
    *ALL_BATCHNORMS,
    nn.LayerNorm,
    nn.InstanceNorm1d,
    nn.InstanceNorm2d,
    nn.InstanceNorm3d,
    nn.GroupNorm,
    nn.LocalResponseNorm,
]

ALL_CONVS = [
    nn.Conv1d,
    nn.Conv2d,
    nn.Conv3d,
    nn.ConvTranspose1d,
    nn.ConvTranspose2d,
    nn.ConvTranspose3d,
]

ALL_LAYERS = [
    nn.Linear,
    *ALL_CONVS,
]


# endregion

# region norm non-affine
def initialize_norms_as_noaffine(m):
    if isinstance(m, ALL_NORMS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 1.)


def initialize_layernorm_as_noaffine(m):
    if isinstance(m, nn.LayerNorm):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 1.)


def initialize_batchnorm_as_noaffine(m):
    if isinstance(m, ALL_BATCHNORMS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 1.)


# endregion

# region norm identity
def initialize_norms_as_identity(m):
    if isinstance(m, ALL_NORMS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 0.)


def initialize_layernorm_as_identity(m):
    if isinstance(m, nn.LayerNorm):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 0.)
    else:
        raise NotImplementedError


def initialize_batchnorm_as_identity(m):
    if isinstance(m, ALL_BATCHNORMS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)
        if m.weight is not None:
            nn.init.constant_(m.weight, 0.)
    else:
        raise NotImplementedError


# endregion

# region bias=0
def initialize_bias_to_zero(m):
    if isinstance(m, ALL_LAYERS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


def initialize_linear_bias_to_zero(m):
    if isinstance(m, nn.Linear):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


def initialize_conv_bias_to_zero(m):
    if isinstance(m, ALL_CONVS):
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


# endregion

# region xavier + bias=0
def initialize_xavier_uniform_zero_bias(m):
    if isinstance(m, ALL_LAYERS):
        nn.init.xavier_uniform_(m.weight)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


# endregion

# region trunc_normal(std=0.02) + bias=0
def initialize_truncnormal_zero_bias(m, std=0.02):
    if isinstance(m, ALL_LAYERS):
        nn.init.trunc_normal_(m.weight, std=std)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


def initialize_linear_truncnormal_zero_bias(m, std=0.02):
    if isinstance(m, nn.Linear):
        nn.init.trunc_normal_(m.weight, std=std)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0.)


# endregion

# region initialize merged nn.Linear layers like seperate nn.Linear layers
def initialize_qkv_seperately(model):
    # https://github.com/facebookresearch/moco-v3/blob/main/vits.py#L35
    for full_name, module in model.named_modules():
        last_name = full_name.split(".")[-1]
        if last_name == "qkv":
            # treat the weights of Q, K, V separately
            val = (6 / (module.weight.shape[0] // 3 + module.weight.shape[1])) ** 0.5
            nn.init.uniform_(module.weight, -val, val)
        if last_name == "qkv_mlpin":
            # treat the weights of Q, K, V and MLP-in separately
            # only implemented for mlp_ratio=4
            input_dim = module.weight.shape[1]
            assert module.weight.shape[0] == 7 * input_dim
            qkv_bound = (3 / input_dim) ** 0.5
            mlpin_bound = (6 / (5 * input_dim)) ** 0.5
            nn.init.uniform_(module.weight[:3 * input_dim], -qkv_bound, qkv_bound)
            nn.init.uniform_(module.weight[3 * input_dim:], -mlpin_bound, mlpin_bound)


def initialize_modulation_seperately(model):
    for full_name, module in model.named_modules():
        last_name = full_name.split(".")[-1]
        if last_name == "modulation":
            # a modulation produces a stack of vectors -> treat each vector seperately
            val = (6 / (module.weight.shape[0] // 2 + module.weight.shape[1])) ** 0.5
            nn.init.uniform_(module.weight, -val, val)


def initialize_seperately(model, name, denominator):
    for full_name, module in model.named_modules():
        last_name = full_name.split(".")[-1]
        if last_name == name:
            val = (6 / (module.weight.shape[0] // denominator + module.weight.shape[1])) ** 0.5
            nn.init.uniform_(module.weight, -val, val)
# endregion



import unittest

import torch
from torch import nn

from kappamodules.layers import WeightNormLinear


class TestWeightNormLinear(unittest.TestCase):
    def is_equal_to_native(self, dim_in, dim_out, bias, fixed_g, seed=49385):
        # native layer
        torch.manual_seed(seed)
        native = nn.utils.weight_norm(nn.Linear(dim_in, dim_out, bias=bias))
        if fixed_g:
            native.weight_g.data.fill_(1)
        # custom layer
        torch.manual_seed(seed)
        custom = WeightNormLinear(dim_in, dim_out, bias=bias, fixed_g=fixed_g)
        # checks
        self.assertTrue(torch.all(native.weight_g == custom.weight_g))
        self.assertTrue(torch.all(native.weight_v == custom.weight_v))
        if bias:
            self.assertTrue(torch.all(native.bias == custom.bias))
        x = torch.randn(2, dim_in, generator=torch.Generator().manual_seed(seed))
        self.assertTrue(torch.all(native(x) == custom(x)))

    def test_is_equal_to_native_nobias_nofixedg(self):
        self.is_equal_to_native(dim_in=4, dim_out=8, bias=False, fixed_g=False)

    def test_is_equal_to_native_bias_nofixedg(self):
        self.is_equal_to_native(dim_in=4, dim_out=8, bias=True, fixed_g=False)

    def test_is_equal_to_native_nobias_fixedg(self):
        self.is_equal_to_native(dim_in=4, dim_out=8, bias=False, fixed_g=True)

    def test_is_equal_to_native_bias_fixedg(self):
        self.is_equal_to_native(dim_in=4, dim_out=8, bias=True, fixed_g=True)



import unittest

import torch
from torch import nn

from kappamodules.layers import DropPath
from original_modules.original_drop_path import OriginalDropPath


class TestEfficientDropPath(unittest.TestCase):
    def is_equal_to_original(self, dim, drop_prob, stochastic_drop_prob, scale_by_keep, training, seed):
        # create layers
        torch.manual_seed(seed)
        custom_sequential_layer = nn.Linear(dim, dim)
        torch.manual_seed(seed)
        custom_standalone_layer = nn.Linear(dim, dim)
        torch.manual_seed(seed)
        original_layer = nn.Linear(dim, dim)
        # create DropPath modules
        kwargs = dict(drop_prob=drop_prob, stochastic_drop_prob=stochastic_drop_prob, scale_by_keep=scale_by_keep)
        custom_sequential = DropPath(custom_sequential_layer, **kwargs)
        custom_standalone = DropPath(**kwargs)
        original = OriginalDropPath(**kwargs)
        # prepare forward
        if not training:
            custom_sequential = custom_sequential.eval()
            custom_standalone = custom_standalone.eval()
            original = original.eval()

        # create data (and make sure that DropPath does nothing inplace)
        x = torch.randn(10, dim, generator=torch.Generator().manual_seed(seed))
        og_x = x.clone()
        # original forward + backward
        torch.manual_seed(seed)
        original_y = x + original(original_layer(x))
        original_y.mean().backward()
        # custom sequential forward
        torch.manual_seed(seed)
        custom_sequential_y = custom_sequential(x)
        custom_sequential_y.mean().backward()
        # custom standalone forward
        torch.manual_seed(seed)
        custom_standalone_y = custom_standalone(x, residual_path=custom_standalone_layer)
        custom_standalone_y.mean().backward()
        # check general
        self.assertTrue(torch.all(x == og_x))
        # check custom_sequential == original
        self.assertTrue(torch.allclose(original_y, custom_sequential_y))
        self.assertTrue(torch.allclose(original_layer.weight.grad, custom_sequential_layer.weight.grad))
        self.assertTrue(torch.allclose(original_layer.bias.grad, custom_sequential_layer.bias.grad))
        # check custom_standalone == original
        self.assertTrue(torch.allclose(original_y, custom_standalone_y))
        self.assertTrue(torch.allclose(original_layer.weight.grad, custom_standalone_layer.weight.grad))
        self.assertTrue(torch.allclose(original_layer.bias.grad, custom_standalone_layer.bias.grad))

    def test_standalone_or_sequential(self):
        layer = nn.Linear(4, 4)
        drop_path = DropPath(layer, drop_prob=0.2)
        with self.assertRaises(AssertionError):
            drop_path(torch.randn(10, 4), residual_path=layer)

    def test_03_stochasticsize_scalebykeep_training(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=True,
            scale_by_keep=True,
            training=True,
            seed=4389,
        )

    def test_03_fixedsize_scalebykeep_training(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=False,
            scale_by_keep=True,
            training=True,
            seed=4389,
        )

    def test_03_stochasticsize_noscalebykeep_training(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=True,
            scale_by_keep=False,
            training=True,
            seed=4389,
        )

    def test_03_stochasticsize_scalebykeep_eval(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=True,
            scale_by_keep=True,
            training=False,
            seed=4389,
        )

    def test_03_fixedsize_noscalebykeep_training(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=False,
            scale_by_keep=False,
            training=True,
            seed=4389,
        )

    def test_03_fixedsize_scalebykeep_eval(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=False,
            scale_by_keep=True,
            training=False,
            seed=4389,
        )

    def test_03_fixedsize_noscalebykeep_eval(self):
        self.is_equal_to_original(
            dim=4,
            drop_prob=0.3,
            stochastic_drop_prob=False,
            scale_by_keep=False,
            training=False,
            seed=4389,
        )



import torch
from torch import nn


class OriginalDropPath(nn.Module):
    """ DropPath as in timm.layers.drop """

    def __init__(self, drop_prob: float = 0., scale_by_keep: bool = True, stochastic_drop_prob: bool = True):
        super().__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep
        self.stochastic_drop_prob = stochastic_drop_prob

    def forward(self, x):
        if self.drop_prob == 0. or not self.training:
            return x
        keep_prob = 1 - self.drop_prob
        shape = (x.shape[0],) + (1,) * (x.ndim - 1)
        if self.stochastic_drop_prob:
            # number of dropped values varies between batches
            random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
        else:
            # number of dropped values is constant between batches
            num_keep = int(len(x) * keep_prob)
            ids_keep = torch.randperm(len(x), device=x.device)[:num_keep]
            random_tensor = torch.zeros(len(x), device=x.device, dtype=x.dtype)
            random_tensor[ids_keep] = 1
            random_tensor = random_tensor.view(*shape)
        if keep_prob > 0.0 and self.scale_by_keep:
            random_tensor.div_(keep_prob)
        return x * random_tensor
